@@ -1,82 +1,20 @@
-"""python simplex implementation"""
+"""Yet another python simplex implementation"""
 
-import argparse
 import ast
+import copy
+import csv
 import dataclasses
 import logging
+import os
 import sys
 import typing
 
-# import numba as nb  # type:ignore
+import jinja2
 import numpy as np
 
+from .args import Argparser
 
-class Argparser:  # pylint: disable=too-few-public-methods
-    """Argparser class."""
-
-    def __init__(self):
-        parser = argparse.ArgumentParser()
-        parser.add_argument(
-            "--equs",
-            "-e",
-            type=str,
-            help="the file containing the equations",
-            default=False,
-        )
-        parser.add_argument(
-            "--slack",
-            "-s",
-            type=str,
-            help="slack variable base name, names are creted"
-            "by adding a number to the string",
-            default="s",
-        )
-        parser.add_argument(
-            "--aux",
-            "-a",
-            type=str,
-            help="aux variable base name, names are creted"
-            "by adding a number to the string",
-            default="xa",
-        )
-        parser.add_argument(
-            "--iter",
-            "-i",
-            type=int,
-            help="maximum number of iterations",
-            default=50,
-        )
-        parser.add_argument(
-            "--min",
-            "-m",
-            action="store_true",
-            help="determines whether its a minimization problem."
-            "if not, its a maximization problem",
-            default=False,
-        )
-        parser.add_argument(
-            "--verbose",
-            "-v",
-            action="store_true",
-            help="whether to print output verbosely",
-            default=False,
-        )
-        parser.add_argument(
-            "--debug",
-            "-d",
-            action="store_true",
-            help="whether to print debug info",
-            default=False,
-        )
-        # TODO- not being used right now
-        parser.add_argument(
-            "--numba",
-            "-n",
-            action="store_true",
-            help="whether to print debug info",
-            default=False,
-        )
-        self.args = parser.parse_args()
+# import numba as nb  # type:ignore
 
 
 @dataclasses.dataclass
@@ -88,22 +26,95 @@ class Equation:
     rhs: float
 
 
-@dataclasses.dataclass
-class LP:
+# pylint: disable=too-few-public-methods
+class LP_Problem:  # pylint: disable=too-many-instance-attributes
     """This class holds the information for an LP problem"""
 
-    equs: typing.Optional[typing.List[Equation]]
-    cost_equ: typing.Optional[Equation]
-    binds: typing.Optional[typing.List[Equation]]
-    A: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    b: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    B: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    B_inv: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    C: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    y_k: typing.Optional[np.ndarray[typing.Any, np.dtype[np.float32]]]
-    basis_column_list: typing.Optional[typing.List[int]]
-    sorted_var_list: typing.Optional[typing.List[str]]
-    is_minimum: typing.Optional[bool]
+    def __init__(self, argparser: Argparser):
+        self.equs: typing.List[Equation] = []
+        self.cost_equ: Equation = Equation({}, "", 0.0)
+        self.binds: typing.List[Equation] = []
+        self.A: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.b: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.B: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.B_inv: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.C: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.y_k: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.basis_column_list: typing.List[int] = []
+        self.var_sorted_list: typing.List[str] = []
+        self.is_minimum: bool = True
+        self.has_identity: bool = False
+        self.var_list: typing.Dict[str, bool] = {}
+        self.basis_is_identity: bool = False
+        self.k: int = 0
+        self.r: int = 0
+        self.w: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.C_b: np.ndarray[typing.Any, np.dtype[np.float32]]
+        self.zj_cj: np.ndarray[typing.Any, np.dtype[np.float32]]
+
+
+def prettify_equs(equ: Equation) -> str:
+    """A custom jinja filter that pretty-prints the equations."""
+    equ_stringified: str = ""
+    # print(equ)
+    for k, v in equ.vars_mults.items():
+        equ_stringified += k + "*" + repr(v) + " "
+    equ_stringified += equ.operand + " " + repr(equ.rhs)
+    return equ_stringified
+
+
+def tabularize_matrix(
+    mat: np.ndarray[typing.Any, np.dtype[np.float32]]
+) -> str:
+    """A custom jinja filter that returns a tabularized."""
+    m: int = mat.shape[0]
+    n: int = mat.shape[1]
+    result: str = "[ \n"
+    for i in range(m):
+        for j in range(n):
+            result += repr(mat[i][j]).ljust(8, " ")
+        result += "\n"
+    result += " ]"
+
+    return result
+
+
+def write_template_head(path: str, lp_problem: LP_Problem):
+    """Write the head of the template to a file."""
+    environment = jinja2.Environment(
+        autoescape=True,
+        loader=jinja2.FileSystemLoader(os.path.join(".", "dsimplex")),
+    )
+    environment.filters["prettify_equs"] = prettify_equs
+    template = environment.get_template(os.path.join("result_head.jinja2"))
+    temp_head = template.render({"lp_problem": lp_problem})
+    with open(path, encoding="utf-8", mode="a+") as out_file:
+        out_file.write(temp_head)
+
+
+def write_round_result(path: str, lp_problem: LP_Problem):
+    """Print the content we have into a file."""
+    environment = jinja2.Environment(
+        autoescape=True,
+        loader=jinja2.FileSystemLoader(os.path.join(".", "dsimplex")),
+    )
+    environment.filters["tabularize_matrix"] = tabularize_matrix
+    template = environment.get_template(os.path.join("result_template.jinja2"))
+    round_result = template.render({"lp_problem": lp_problem})
+    with open(path, encoding="utf-8", mode="a+") as out_file:
+        out_file.write(round_result)
+
+
+def write_template_tail(path: str, lp_problem: LP_Problem, result: float):
+    """Print the content we have into a file."""
+    environment = jinja2.Environment(
+        autoescape=True,
+        loader=jinja2.FileSystemLoader(os.path.join(".", "dsimplex")),
+    )
+    template = environment.get_template(os.path.join("result_tail.jinja2"))
+    temp_tail = template.render({"lp_problem": lp_problem, "result": result})
+    with open(path, encoding="utf-8", mode="a+") as out_file:
+        out_file.write(temp_tail)
 
 
 @typing.no_type_check
@@ -234,7 +245,7 @@ def expr_visitor(
     return None, equ
 
 
-def parse_equations(equ_file: str, debug: bool) -> typing.List[Equation]:
+def parse_equ_py(equ_file: str, debug: bool) -> typing.List[Equation]:
     """Parses the equations as pyithon expressions."""
     equs = []
     with open(equ_file, encoding="utf-8") as equ_expr:
@@ -260,6 +271,39 @@ def parse_equations(equ_file: str, debug: bool) -> typing.List[Equation]:
     return equs
 
 
+def parse_equ_csv(csv_file: str, delim: str) -> typing.List[Equation]:
+    """Read in a CSV file and convert it to the common format"""
+    equ_list = []
+    equ = Equation({}, "", 0.0)
+    with open(csv_file, newline="", encoding="utf-8") as csvfile:
+        reader = csv.reader(csvfile, delimiter=delim, quotechar="|")
+        names = next(reader)
+        for row in reader:
+            # print(row)
+            if row[-1] != "":
+                equ.rhs = float(row[-1])
+            if row[-2] != "":
+                equ.operand = row[-2]
+            for i, v in enumerate(row[:-2]):
+                if v != "":
+                    equ.vars_mults[names[i]] = float(v)
+            equ_list.append(copy.deepcopy(equ))
+            equ.operand = ""
+            equ.vars_mults = {}
+            equ.rhs = 0.0
+
+    return equ_list
+
+
+def parse_equ(argparser: Argparser) -> typing.List[Equation]:
+    """Driver for the equation parsers"""
+    if argparser.args.equs:
+        return parse_equ_py(argparser.args.equs, argparser.args.debug)
+    if argparser.args.csv:
+        return parse_equ_csv(argparser.args.csv, argparser.args.delim)
+    return []
+
+
 def add_slack_vars(
     equs: typing.List[Equation], slack: str, verbose: bool
 ) -> typing.Tuple[
@@ -279,7 +323,6 @@ def add_slack_vars(
     for i, equ in enumerate(equs):
         # we assume that b > 0, if not we multiply the equation
         # by -1 and flip the operand accordingly
-        # TODO-test me
         if equ.rhs < 0:
             for j in equ.vars_mults:
                 equ.vars_mults[j] = equ.vars_mults[j] * -1
@@ -422,30 +465,19 @@ def build_abc(
     return A, b, C, var_sorted_list
 
 
-def construct_lp_problem(
-    equs: typing.List[Equation],
-    verbose: bool,
-    slack: str,
-    is_minimum: bool,
-    aux_var_name: str,
-) -> typing.Tuple[
-    np.ndarray[typing.Any, np.dtype[np.float32]],
-    np.ndarray[typing.Any, np.dtype[np.float32]],
-    np.ndarray[typing.Any, np.dtype[np.float32]],
-    np.ndarray[typing.Any, np.dtype[np.float32]],
-    typing.List[Equation],
-    Equation,
-    typing.List[str],
-    typing.Dict[str, bool],
-    bool,
-    typing.List[int],
-]:
+def construct_lp_problem(lp_problem: LP_Problem, argparser: Argparser) -> None:
     """Construct the LP problem."""
+    verbose = argparser.args.verbose
+    slack = argparser.args.slack
+    is_minimum = argparser.args.min
+    aux_var_name = argparser.args.aux
+    equs = lp_problem.equs
+
     var_list: typing.Dict[str, bool] = {}
     var_sorted_list: typing.List[str] = []
     slack_counter: int = 0
 
-    equs, cost_equ, _, slack_counter, var_list = add_slack_vars(
+    equs, cost_equ, binds, slack_counter, var_list = add_slack_vars(
         equs, slack, verbose
     )
 
@@ -459,15 +491,14 @@ def construct_lp_problem(
 
     has_identity, col_list = find_identity(A)
     if has_identity and col_list is not None:
-        print("col_list:", col_list)
+        # print("col_list:", col_list)
         for _, v in col_list.items():
             col_list_list.append(v)
     else:
         # big M
         m_zero: float = calculate_big_m_zero(A, b, C)
-        print("m_zero:", m_zero)
+        # print("m_zero:", m_zero)
         ones_column_list = get_ones(A)
-        print("ones_column_list:\n", ones_column_list)
         build_identity(
             A,
             equs,
@@ -491,18 +522,18 @@ def construct_lp_problem(
     # we will always have an identity basis
     B = np.identity(m, dtype=np.float32)
 
-    return (
-        A,
-        b,
-        B,
-        C,
-        equs,
-        cost_equ,
-        var_sorted_list,
-        var_list,
-        has_identity,
-        col_list_list,
-    )
+    lp_problem.A = A
+    lp_problem.b = b
+    lp_problem.B = B
+    lp_problem.C = C
+    lp_problem.equs = equs
+    lp_problem.cost_equ = cost_equ
+    lp_problem.A = A
+    lp_problem.var_sorted_list = var_sorted_list
+    lp_problem.var_list = var_list
+    lp_problem.has_identity = has_identity
+    lp_problem.basis_column_list = col_list_list
+    lp_problem.binds = binds
 
 
 # @nb.jit(nopython=True, cache=True)
@@ -539,7 +570,6 @@ def find_identity(
         ones_count = 0
         last_one_row = 0
 
-    print("ones:", ones)
     if ones == (m * (m + 1)) / 2:
         return True, col_list
     return False, col_list
@@ -560,7 +590,6 @@ def get_ones(
     # bye-bye cache locality
     for j in range(0, n):
         for i in range(0, m):
-            # col_sum += col_sum + A[i, j]
             if A[i, j] == 0:
                 continue
             if A[i, j] == 1:
@@ -572,13 +601,11 @@ def get_ones(
             else:
                 seen_trash = True
                 break
-        # if seen_a_one and col_sum == 1 and not seen_trash:
         if seen_a_one and not seen_trash:
             col_list[one_index] = j
         seen_a_one = False
         seen_trash = False
         one_index = -1
-        # col_sum = 0
 
     return col_list
 
@@ -662,9 +689,9 @@ def calculate_objective(
         B_inv = B
     else:
         B_inv = invert_matrix(B)
-    print("B_inv:\n", B_inv)
+    # print("B_inv:\n", B_inv)
 
-    print("basic_var_column_list: ", basic_var_column_list)
+    # print("basic_var_column_list: ", basic_var_column_list)
     for i, v in enumerate(basic_var_column_list):
         C_b[0, i] = C[0, v]
 
@@ -681,27 +708,6 @@ def calculate_objective(
     return B_inv, objectives, w, C_b
 
 
-# NOTE- there could be more than one minimum
-# also this does not prevent cycling among extreme points
-# currently unused
-def get_non_negative_min(
-    M,
-    y_k: np.ndarray[typing.Any, np.dtype[np.float32]],
-) -> typing.Tuple[float, int]:
-    """Get the index of the leaving variable in the basis var list.
-    This needs to get translated into the actual index of the variable."""
-    n = M.shape[0]
-    minimum: float = 1e9
-    minimum_index: int = 0
-    for i in range(0, n):
-        if y_k[i, 0] > 0:
-            if M[i, 0] < minimum:
-                minimum = M[i, 0]
-                minimum_index = i
-
-    return minimum, minimum_index
-
-
 def get_leaving_var_lexi(
     M,
     B_inv,
@@ -786,7 +792,7 @@ def determine_leaving(
         print("b_bar:\n", b_bar)
 
     b_bar_div_y = np.divide(b_bar[:, :], y_k)
-    print("b_bar_div_y:\n", b_bar_div_y)
+    # print("b_bar_div_y:\n", b_bar_div_y)
     _, r = get_leaving_var_lexi(b_bar_div_y, B_inv, A, y_k)
     rr = basis_col_list[r]
     if verbose:
@@ -846,12 +852,12 @@ def calculate_optimal(
     optim: float = 0.0
 
     x_b = np.dot(B_inv, b)
-    print("x_b:\n", x_b)
-    print("optimal solution point:")
+    # print("x_b:\n", x_b)
+    # print("optimal solution point:")
     for basis in basis_col_list:
         print(var_sorted_list[basis])
     Z = np.dot(C_b, x_b)
-    print("Z:\n", Z)
+    # print("Z:\n", Z)
 
     optim = Z[0, 0]
 
@@ -859,16 +865,19 @@ def calculate_optimal(
 
 
 def solve_normal_simplex(
-    A,
-    b,
-    C,
-    B: np.ndarray[typing.Any, np.dtype[np.float32]],
-    basis_is_identity: bool,
-    basis_col_list: typing.List,
-    var_sorted_list: typing.List[str],
+    lp_problem: LP_Problem,
     argparser: Argparser,
-) -> None:
+) -> typing.Optional[float]:
     """Solve using the normal simplex method."""
+    write_template_head(argparser.args.out, lp_problem)
+    A = lp_problem.A
+    b = lp_problem.b
+    C = lp_problem.C
+    B = lp_problem.B
+    basis_is_identity = lp_problem.basis_is_identity
+    basis_col_list = lp_problem.basis_column_list
+    var_sorted_list = lp_problem.var_sorted_list
+
     verbose: bool = argparser.args.verbose
     round_count: int = 0
     while True:
@@ -890,32 +899,30 @@ def solve_normal_simplex(
         if argparser.args.min:
             if extrmem_zj_cj < 0:
                 # we are done
-                print(
-                    "optimal min is:",
-                    calculate_optimal(
-                        b, B_inv, C_b, basis_col_list, var_sorted_list
-                    ),
+                opt: float = calculate_optimal(
+                    b, B_inv, C_b, basis_col_list, var_sorted_list
                 )
-                break
+                print("optimal min is:", opt)
+                write_template_tail(argparser.args.out, lp_problem, opt)
+                return opt
         else:
             if extrmem_zj_cj > 0:
                 # we are done
-                print(
-                    "optimal max is:",
-                    calculate_optimal(
-                        b, B_inv, C_b, basis_col_list, var_sorted_list
-                    ),
+                opt = calculate_optimal(
+                    b, B_inv, C_b, basis_col_list, var_sorted_list
                 )
-                break
+                print("optimal max is:", opt)
+                write_template_tail(argparser.args.out, lp_problem, opt)
+                return opt
 
         y_k = np.dot(B_inv, A[:, k : k + 1])
-        print("y_k:\n", y_k)
+        # print("y_k:\n", y_k)
         if np.all(np.less_equal(y_k, 0)):
             # we are done
             # TODO- print the direction along which the value is unbounded
             print("unbounded optimal value.")
-            break
-        print(y_k)
+            return None
+        # print(y_k)
         r, B = determine_leaving(
             k, A, B, B_inv, b, y_k, basis_col_list, basis_is_identity, verbose
         )
@@ -927,44 +934,121 @@ def solve_normal_simplex(
                 leaving_col = i
         basis_col_list[leaving_col] = k
 
-        if verbose:
-            print("B:\n", B)
-        print("-------------------------------------------------")
+        # if verbose:
+        #     print("B:\n", B)
+        # print("-------------------------------------------------")
         if round_count > argparser.args.iter:
             print("too many iterations.")
-            break
+            return None
+
+        lp_problem.A = A
+        lp_problem.B = B
+        lp_problem.y_k = y_k
+        lp_problem.B_inv = B_inv
+        lp_problem.r = r
+        lp_problem.k = k
+        lp_problem.b = b
+        lp_problem.w = w
+        lp_problem.C_b = C_b
+        lp_problem.zj_cj = objectives
+        write_round_result(argparser.args.out, lp_problem)
 
 
-def dsimplex() -> None:
+def dsimplex() -> typing.Optional[float]:
     """The entry point for the module."""
     argparser = Argparser()
-    verbose = argparser.args.verbose
-    equs = parse_equations(argparser.args.equs, argparser.args.debug)
-    (
-        A,
-        b,
-        B,
-        C,
-        equs,
-        _,
-        var_sorted_list,
-        _,
-        basis_is_identity,
-        basis_col_list,
-    ) = construct_lp_problem(
-        equs,
-        verbose,
-        argparser.args.slack,
-        argparser.args.min,
-        argparser.args.aux,
-    )
-    solve_normal_simplex(
-        A,
-        b,
-        C,
-        B,
-        basis_is_identity,
-        basis_col_list,
-        var_sorted_list,
+    lp_problem = LP_Problem(argparser)
+
+    lp_problem.equs = parse_equ(argparser)
+    if len(lp_problem.equs) == 0:
+        print("could not parse the given equations")
+        return None
+
+    construct_lp_problem(lp_problem, argparser)
+
+    return solve_normal_simplex(
+        lp_problem,
         argparser,
     )
+
+
+def dsimplex_gui(argparser: Argparser) -> typing.Optional[float]:
+    """The entry point for the gui."""
+    lp_problem = LP_Problem(argparser)
+
+    lp_problem.equs = parse_equ(argparser)
+    if len(lp_problem.equs) == 0:
+        print("could not parse the given equations")
+        return None
+
+    construct_lp_problem(lp_problem, argparser)
+
+    return solve_normal_simplex(
+        lp_problem,
+        argparser,
+    )
+
+
+def dsimplex_gui_loop(
+    argparser: Argparser, equs: typing.List[Equation]
+) -> typing.Optional[float]:
+    """The entry point for the gui"""
+    lp_problem = LP_Problem(argparser)
+
+    lp_problem.equs = equs
+    if len(lp_problem.equs) == 0:
+        print("could not parse the given equations")
+        return None
+
+    construct_lp_problem(lp_problem, argparser)
+
+    return solve_normal_simplex(
+        lp_problem,
+        argparser,
+    )
+
+
+def isfloat(num):
+    try:
+        float(num)
+        return True
+    except ValueError:
+        return False
+
+
+def parse_equ_csv_loop(
+    argparser: Argparser,
+) -> typing.List[typing.Optional[float]]:
+    """Solves in a loop."""
+    equ_list: typing.List = []
+    equ = Equation({}, "", 0.0)
+    result: typing.List = []
+    with open(argparser.args.csv, newline="", encoding="utf-8") as csvfile:
+        reader = csv.reader(
+            csvfile, delimiter=argparser.args.delim, quotechar="|"
+        )
+        names = next(reader)
+        for row in reader:
+            if row[0] == "null":
+                result.append(dsimplex_gui_loop(argparser, equ_list))
+                equ_list = []
+                continue
+            if not any(isfloat(row_element) for row_element in row):
+                result.append(dsimplex_gui_loop(argparser, equ_list))
+                equ_list = []
+                names = row
+                continue
+            if row[-1] != "":
+                equ.rhs = float(row[-1])
+            if row[-2] != "":
+                equ.operand = row[-2]
+            for i, v in enumerate(row[:-2]):
+                if v != "":
+                    equ.vars_mults[names[i]] = float(v)
+            equ_list.append(copy.deepcopy(equ))
+            equ.operand = ""
+            equ.vars_mults = {}
+            equ.rhs = 0.0
+
+        result.append(dsimplex_gui_loop(argparser, equ_list))
+        return result
